/*
 *  Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package javax.crypto;

import java.security.InvalidKeyException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;

/**
 * The <i>Service Provider Interface</i> (<b>SPI</b>) definition for the {@code
 * SecretKeyFactory} class.
 * 
 * @since Android 1.0
 */
public abstract class SecretKeyFactorySpi {

    /**
     * Creates a new {@code SecretKeyFactorySpi} instance.
     * @since Android 1.0
     */
    public SecretKeyFactorySpi() {
    }

    /**
     * Generate a secret key from the specified key specification.
     * 
     * @param keySpec
     *            the key specification.
     * @return a secret key.
     * @throws InvalidKeySpecException
     *             if the specified key specification cannot be used to generate
     *             a secret key.
     * @since Android 1.0
     */
    protected abstract SecretKey engineGenerateSecret(KeySpec keySpec)
            throws InvalidKeySpecException;

    /**
     * Returns the key specification of the specified secret key.
     * 
     * @param key
     *            the secret key to get the specification from.
     * @param keySpec
     *            the target key specification class.
     * @return an instance of the specified key specification class.
     * @throws InvalidKeySpecException
     *             if the specified secret key cannot be transformed into the
     *             requested key specification.
     * @since Android 1.0
     */
    @SuppressWarnings("unchecked")
    protected abstract KeySpec engineGetKeySpec(SecretKey key, Class keySpec)
            throws InvalidKeySpecException;

    /**
     * Translates the specified secret key into an instance of the corresponding
     * key from the provider of this key factory.
     * 
     * @param key
     *            the secret key to translate.
     * @return the corresponding translated key.
     * @throws InvalidKeyException
     *             if the specified key cannot be translated using this key
     *             factory.
     * @since Android 1.0
     */
    protected abstract SecretKey engineTranslateKey(SecretKey key)
            throws InvalidKeyException;
}